# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing


def fix_variables(*args, **kwargs) -> typing.NoReturn:
    """Removed

    .. deprecated:: 0.10.0

        This function was migrated to
        `dwave-preprocessing <https://github.com/dwavesystems/dwave-preprocessing>`_

        Use ``dwave.preprocessing.roof_duality`` instead.

    """
    raise TypeError(
        "fix_variables() was moved to dwave-preprocessing in dimod 0.10.0. "
        "Install dwave-preprocessing and use dwave.preprocessing.roof_duality instead. "
        "You can do so with 'pip install dwave-preprocessing'. "
        "This stub function will be removed in dimod 0.13.0.",
        )
